//  -*- C++ -*-
//    File: TSWindowMod.hh
//   Author: David Chin <dwchin@umich.edu> +1-734-730-1274
// $Id: TSWindowMod.hh,v 1.2 2009/02/13 05:29:35 keithr Exp $
#ifndef LOCKLOSS_TSWINDOWMOD_HH
#define LOCKLOSS_TSWINDOWMOD_HH

#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <deque>
#include <stdexcept>
#include <iterator>
#include <cstdlib>

#include "Interval.hh"
#include "TSeries.hh"

typedef unsigned long ulong_t;

//
// We would like to serve X hours of N second resolution data to the
// DMT viewer.  So, I want a queue-like object with an upper limit on
// length which acts like a window on a time series.
//
// All Time and Interval objects have the nanoseconds field dropped.
//
template <class T> 
class TSWindowMod
{
public:
    /** @memo Default constructor
     *  @param len   Maximum length of window
     *  @param dt    Sampling interval. NOTE: nanoseconds part is ignored
     *  @param dbg   Debug level
     *  @param name  Name of this {\tt TSWindowMod}. NOTE: no whitespace
     *  @param savefilename Name of savefile,
     *  @param dummyval Value to use as dummy padding
     */
    TSWindowMod(T                   dummyval,
             size_t              len  = 8192,
             const Interval    & dt   = Interval(1,0),
             int                 dbg  = 0,
             const std::string & name = std::string("TSWindow_noname"),
             const std::string & savefilename = std::string())
        : mDebug(dbg),
          mName(name),
          mStartTime(0),
          mDt(dt.GetS()),
          mData(0),
          mSaveFileName(savefilename),
          mDummyVal(dummyval)
        {
            // search for "space" or "tab" in name
            if (mName.find(" ") != std::string::npos ||
                mName.find("\t") != std::string::npos) {
                std::string eObj("TSWindowMod::TSWindowMod(): ERROR: name cannot contain whitespace");
                std::cerr << eObj << std::endl;
                throw std::runtime_error(eObj);
            }

            // length must be > 0
            if (len <= 0) {
                std::string eObj("TSWindowMod::TSWindowMod(): ERROR: max length must be >=0");
                std::cerr << eObj << std::endl;
                throw std::runtime_error(eObj);
            }

            setMaxLength(len);

            if (mDebug > 2)
                dump();
        }
    
// UNUSED FOR NOW. MAYBE NEVER!!
    /** @memo Initializer
     *  @param len  Maximum length of window
     *  @param starttime Start time of window
     *  @param dt   Sampling interval. NOTE: nanoseconds part is ignored
     *  @param dbg  Debug level
     *  @param name Name of this {\tt TSWindowMod}
     */
//      void init(T                   dummyval,
//                const std::string & savefilename,
//                size_t              len  = 8192,
//                const Time        & starttime = Time(),
//                const Interval    & dt   = Interval(1,0),
//                int                 dbg  = 0,
//                const std::string & name = std::string("TSWindow_noname"))
//          {
//              mDebug = dbg;
//              mName  = name;
//              mStartTime = starttime.getS();
//              mDt        = dt.GetS();
//              mData      = 0;
//              mSaveFileName = savefilename;
//              mDummyVal  = dummyval;

//              if (mName.find(" ") != std::string::npos ||
//                  mName.find("\t") != std::string::npos) {
//                  std::string eObj("TSWindowMod::TSWindowMod(): ERROR: name cannot contain whitespace");
//                  std::cerr << eObj << std::endl;
//                  throw std::runtime_error(eObj);
//              }

//              // length must be > 0
//              if (len <= 0) {
//                  std::string eObj("TSWindowMod::TSWindowMod(): ERROR: maxlength must be >=0");
//                  std::cerr << eObj << std::endl;
//                  throw std::runtime_error(eObj);
//              }

//              setMaxLength(len);
//          }
   
    /// Destructor
    ~TSWindowMod()  { delete [] mData; }
   
    /// max. length of window
    size_t maxLength() const { return  mMaxLength; }
   
    /// cur. length of window
    size_t length() const { return mQueue.size(); }
   
    /// start time
    Time startTime() const { return Time(mStartTime,0); }
    Time t0() const { return Time(mStartTime,0); }
   
    /// sampling interval
    Interval dt() const { return Interval(mDt,0); }

    /// dummy value
    T dummyVal() const { return mDummyVal; }
   
    /// Name of this TSWindowMod
    std::string name() const { return mName; }
   
    /// Savefile name
    std::string saveFileName() const { return mSaveFileName; }
   
    /// Print info
    void printInfo() const
        {
            std::cout << "TSWindowMod: name = " << mName << std::endl;
            std::cout << "            dt = " << mDt << std::endl;
            std::cout << "    max length = " << mMaxLength << std::endl;
            std::cout << "    start time = " << mStartTime << std::endl;
        }

   
    /// set name
    void setName(const std::string & name)
        {
            if (mDebug > 0)
                std::cout << "TSWindowMod: setting name: " << name << std::endl;
    
            mName = name;
        }
   
    /// set interval
    void setDt(const Interval & dt)
        {
            if (mDebug > 0)
                std::cout << "TSWindowMod: setting interval: " << dt << std::endl;
    
            mDt = dt.GetS();
        }
    
    void setInterval(const Interval & dt) { setDt(dt); }
   
    /// set start time
    void setStartTime(const Time & t0)
        {
            if (mDebug > 0)
                std::cout << "TSWindowMod: setting start time: " << t0 << std::endl;
    
            mStartTime = t0.getS();
        }

   
    /// set max length -- side effects: sets mMaxLength, allocates data array
    void setMaxLength(size_t len = 8192)
        {
            if (mDebug > 0)
                std::cout << "TSWindowMod: setting max length: " << len << std::endl;

            if (len <= 0) {
                std::string eObj("TSWindowMod::setMaxLength(): ERROR: max length must be >=0");
                std::cerr << eObj << std::endl;
                throw std::runtime_error(eObj);
            }
    
            mMaxLength = len;

            // allocate space for data array
    
            if (mData != (T *)NULL) {
                if (mDebug > 0)
                    std::cout << "TSWindowMod: mData != NULL" << std::endl;
        
                delete [] mData;
            }
    
            mData = new T[mMaxLength];

            if (mDebug > 0) {
                std::cout << "TSWindowMod::setMaxLength(): data arrray allocated; length = "
                          << length()
                          << "; dt = " << dt() << std::endl;
            }
        }

   
    /// set debug
    void setDebug(int dbg) { mDebug = dbg; }
      
    /// set savefile name
    void setSaveFileName(const std::string & savefilename)
        {
            mSaveFileName = savefilename;
        }
   

     /// KR - Added new function 20 Feb 2006

     T getLatest()
       {
	 T val;
	 val = -1;
	 if (mQueue.size() > 0 ) {
	   typename std::deque<T>::const_iterator iter = mQueue.end();
	   iter--;
	   val = *iter;	
	 }
	 return val;
       }

     /// KR - Added new function 23 June 2008 (get mean of latest nval values)

     T getMean(int nval)
       {
	 T val, meanval, sumval;
	 sumval = 0.;
	 meanval = -1.;
	 if (mQueue.size() > (unsigned int)(nval+1) ) {
	   typename std::deque<T>::const_iterator iter = mQueue.end();
	   for (int i=0; i<nval; i++) {
	     iter--;
	     val = *iter;
	     sumval+= val;
	   }
	   meanval = sumval / nval;
	 }
	 return meanval;
       }

     /// KR - Added new function 24 June 2008 (get mean of latest nval non-negative values)

     T getMeanPos(int nval)
       {
	 T val, meanval, sumval;
	 meanval = -1.;
	 sumval = 0.;
	 if (mQueue.size() > (unsigned int)(nval+1) ) {
	   typename std::deque<T>::const_iterator iter = mQueue.end();
	   int i = 0;
	   while (i<nval & iter!=mQueue.begin()) {
	     iter--;
	     val = *iter;
	     if (val>-1.e-6) {
	       sumval += val;
	       i++;
	     }
	   }
	   if (i==nval) meanval = sumval / nval;
	 }
	 return meanval;
       }

    /// Append a value into window
    void append(T val, const Time & t)
        {
            if (mDebug > 1)
                std::cout << "TSWindowMod::append: "
                          << val << " at time t = " << t.getS() << std::endl;
      
            if (mMaxLength == 0) {
                std::cerr << "TSWindowMod: maxLength has not been set" << std::endl;
                std::string eObj = "maxLength is zero";
                throw std::runtime_error(eObj);
            }
      
            // check for illegal input
            if (t.getN() != 0) {
                std::string eObj("TSWindowMod::append(): ERROR: time t has non-zero nanosec field.");
                std::cerr << eObj << std::endl;
                throw std::runtime_error(eObj);
            }
      
            if (t == Time()) {
                std::string eObj("TSWindowMod::append(): ERROR: time t is 0.");
                std::cerr << eObj << std::endl;
                throw std::runtime_error(eObj);
            }
      
            // set start time if it hasn't been set
            if (mStartTime == 0)
                mStartTime = t.getS();
     
            // pad with dummy data if necessary
            if (mQueue.size() > 0 ) {

                // gap is time diff. between last data in queue and
                // data that's about to be appended
                long long gap = computeGap(t);
         
                if (mDebug > 2) {
                    std::cout << "TSWindowMod::append(): gap = " << gap
                              << std::endl
                              << "    mQueue.size() = " << mQueue.size()
                              << std::endl
                              << "    mDt = " << mDt
                              << std::endl
                              << "    t.getS()   = " << t.getS()
                              << std::endl;
                    if (gap > 0)
                        std::cout << "PADDING WITH DUMMY..." << std::endl;
                }
    
                if (gap < 0) {
                    std::string eObj("TSWindowMod::append(): ERROR: given time too early -- can only append to end of current data");
                    std::cerr << eObj << std::endl;
                    throw std::runtime_error(eObj);
                }
    
                while (gap > 0) {
                    smart_push_back(mDummyVal);
                    --gap;
                }
            }

            smart_push_back(val);
        }
   
    /** @memo Fill window with given data, starting with time. Just
     *  overwrites current data. Returns no. of data
     *  filled.
     *  @param startTime Starting time for the TSWindowMod
     *  @param data Vector containing data
     *  @return Number of elements filled.
     */
    size_t fill(const Time & startTime, const std::vector<T> & data)
        {
            if (mDebug > 2) {
                std::cout << "TSWindowMod: fill():     startTime = "
                          << startTime << std::endl;
                std::cout << "TSWindowMod: fill(): i) mStartTime = "
                          << mStartTime << std::endl;
            }

            // check for error conditions
            if (startTime == Time()) {
                std::string eObj("TSWindowMod::fill(): ERROR: given start time is 0.");
                std::cerr << eObj << std::endl;
                throw std::runtime_error(eObj);
            }

            if (startTime.getN() != 0) {
                std::string eObj("TSWindowMod::fill(): ERROR: startTime has non-zero nanosec field");
                std::cerr << eObj << std::endl;
                throw std::runtime_error(eObj);
            }

            size_t numwritten=0;
            
            if (data.size() > mMaxLength) {
                std::cerr << "TSWindowMod::fill(): WARNING: too much data given, truncating" << std::endl;

                mQueue.clear();
                for (numwritten = 0; numwritten < mMaxLength; ++numwritten)
                    mQueue.push_back(data[numwritten]);
            }
            
            if (data.size() <= mMaxLength) {

                mQueue.clear();
                size_t ds = data.size();
                for (numwritten = 0; numwritten < ds; ++numwritten)
                    mQueue.push_back(data[numwritten]);

                if (data.size() < mMaxLength) {
                    std::cerr << "TSWindowMod::fill(): WARNING: too little data given, padding" << std::endl;
                    for (; numwritten < mMaxLength; ++numwritten)
                        mQueue.push_back(mDummyVal);
                }
            }

            mStartTime = startTime.getS();
         
            if (mDebug > 2)
                std::cout << "TSWindowMod: fill(): ii) mStartTime = "
                          << mStartTime << std::endl;
      
            return numwritten;
        }

   
    /** @memo Fill window with nData times the given val, with starting time
     *    startTime
     *  @param numval Number of data elements to fill
     *  @param val    Value to be used for filling
     *  @param startTime Starting time for the TSWindowMod
     *  @return Number of elements filled
     */
    size_t fill(const Time & startTime,
                size_t numval = 0,
                T val = T())
        {
            if (numval == 0)
                numval = mMaxLength;
    
            if (val == T())
                val = mDummyVal;
    
            std::vector<T> padding(numval, val);
            return fill(startTime, padding);
        }
   
    /// fill only if startTime has not been set 
    size_t fillMaybe(const Time & startTime, size_t numval = 0, T val = T())
        {
            if (mStartTime == 0)  {
                if (mDebug > 0) {
                    std::cout << "TSWindowMod::fillMaybe(): mStartTime == 0" << std::endl;
                }
                return fill(startTime, numval, val);
            } else {
                if (mDebug > 0) {
                    std::cout << "TSWindowMod::fillMaybe(): mStartTime == " 
                              << mStartTime << std::endl;
                }
                return mMaxLength;
            }
        }

    /// Fill in given TSeries with data
    void fillTSeries(TSeries &ts) const
        {
            // if interval is 0
            if (mDt == 0) {
                std::string eObj = "TSWindowMod: ERROR: Sampling interval is 0";
                std::cerr << eObj << std::endl;
                throw std::runtime_error(eObj);
            }
      
            // dunno what to do if time is not set right, or dt is
            // not set right.
            if (mStartTime == 0) {
                std::string eObj = "TSWindowMod: ERROR: Start time is 0";
                std::cerr << eObj << std::endl;
                throw std::runtime_error(eObj);
            }
      
            size_t i = 0;
            typename std::deque<T>::const_iterator iter = mQueue.begin();
            for (; iter != mQueue.end(); ++iter, ++i) {
                mData[i] = *iter;
            }
      
            ts.setData(Time(mStartTime,0), Interval(mDt,0), mData, length());
        }

   
    /// Alias for fillTSeries()
    void tseries(TSeries &ts) const
        {
            fillTSeries(ts);
        }
   
    /** Fill in given array with num data; return no. of elements
      * if the num parameter is given, data is filled by most recent
      * num entries.
      */
    size_t fillVector(std::vector<T> &d, size_t num = 0) const
        {
            int i = (size_t)0;
            
            if (num < 0) { // shouldn't happen
                std::string eObj("TSWindowMod::fillVector(): ERROR: num < 0");
                eObj += "; ";
                eObj += mName;
                std::cerr << eObj << std::endl;
                throw std::runtime_error(eObj);
            } else if (num > maxLength()) {
                std::string eObj("TSWindowMod::fillVector(): ERROR: num > maxLength");
                eObj += "; ";
                eObj += mName;
                std::cerr << eObj << std::endl;
                throw std::runtime_error(eObj);
            } else if (num == 0) {
                // make sure there is enough space in the vector
                if (d.size() < length())
                    d.resize(length());
      
                typename std::deque<T>::const_iterator iter = mQueue.begin();
                for (; iter != mQueue.end(); ++iter, ++i)
                    d[i] = *iter;
                
            } else if (num >= 1) {
                if (d.size() < num)
                    d.resize(num);

                i = num-1;
                typename std::deque<T>::const_reverse_iterator iter = mQueue.rbegin();
                for (; iter != mQueue.rend() && i >= 0; ++iter, --i)
                    d[i] = *iter;
            }
      
            return i;
        }


    /// Alias for fillVector
    size_t data(std::vector<T> &d) const
        {
            return fillVector(d);
        }
   
    /** Save data to file. Format:
     *  startTime
     *  data[0]
     *  data[1]
     *  ...
     *  data[N]
     */
    void save()
        {
            if (mSaveFileName.size() == 0) {
                std::string eObj =
                    "TSWindowMod: ERROR: savefile name has not been set";
                std::cerr << eObj << std::endl;
                throw std::runtime_error(eObj);
            } else {

                if (mDebug > 0) 
                    std::cout << "TSWindowMod::save(): attempting to "
                              << "save to " << mSaveFileName << std::endl;

		// bleah. the iostream library in gcc-3.3 does not reset
		// status bits on close().
         
                if (!mSaveFile.is_open()) {
                    if (mDebug > 0) 
                        std::cout << "TSWindowMod::save(): saveFile not open" 
                                  << std::endl;
                    mSaveFile.open(mSaveFileName.c_str(), std::ios_base::out);
                    mSaveFile.clear();  // reset mSaveFile to good state
                    mSaveFile.seekp(std::ios_base::beg);
                }
         
                if (mSaveFile.fail()) {
                    std::string eObj =
                        "TSWindowMod: ERROR: savefile could not be opened";
                    std::cerr << eObj << std::endl;
                    throw std::runtime_error(eObj);
                }

                mSaveFile.setf(std::ios_base::scientific);
                mSaveFile.precision(18);         // overkill, prolly
                mSaveFile << mDebug
                          << " " << mName
                          << " " << mDummyVal
                          << " " << mMaxLength
                          << " " << mQueue.size()
                          << " " << mStartTime
                          << " " << mDt << std::endl;
         
                // Ya know, this will not work except for the most
                // simple of template contents. Or if the class
                // contained has << and >> operators defined. D'oh.
                std::copy(mQueue.begin(), mQueue.end(),
                          std::ostream_iterator<T>(mSaveFile, " "));
         
                mSaveFile.close();

                if (mDebug > 0) 
                    std::cout << "TSWindowMod::save(): save to "
                              << mSaveFileName << " completed"
                              << std::endl;
            }
        }


    /// Restore saved data if available
    void restore()
        {
            if (mDebug > 0) 
                std::cout << "TSWindowMod::restore(): attempting to "
                          << "restore from " << mSaveFileName << std::endl;
                
            if (mSaveFileName.size() == 0) {
                std::string eObj("TSWindowMod::restore(): ERROR: savefile name has not been set");
                std::cerr << eObj << std::endl;
                throw std::runtime_error(eObj);
            } else {
         
                if (!mSaveFile.is_open()) {
                    mSaveFile.open(mSaveFileName.c_str(), std::ios_base::in);
                    mSaveFile.seekg(std::ios_base::beg);

                    if (mSaveFile.fail()) {
                        std::cerr << "TSWindowMod::restore(): WARNING: save file "
                                  << "does not exist; cannot restore" << std::endl;
                        mSaveFile.close();
                        return;
                    }
                }

                // tmp vars to store start time, and timestep
                ulong_t number_saved;  // number of data points that were saved
                char tmpname[128];  // trivial error-check
                mSaveFile >> mDebug >>  tmpname >> mDummyVal
                          >> mMaxLength >> number_saved >> mStartTime >> mDt;

                if (mDebug > 2) {
                    std::cout << mDebug << "\t"
                              << mDummyVal << "\t"
                              << tmpname << "\t"
                              << mMaxLength << "\t"
                              << number_saved << "\t"
                              << mStartTime << "\t"
                              << mDt << std::endl;

                    std::cout << "mName = " << mName << std::endl;
                    std::cout << "mName.size() = " << mName.size() << std::endl;
                    std::cout << "tmpname = " << tmpname << std::endl;
                }

                if (mName.size() > 0 && std::string(tmpname) != mName) {
                    std::cerr << "TSWindowMod::restore(): stored data "
                              << "has different name: expected " << mName
                              << " got " << tmpname << std::endl;
                    std::cerr << "    removing file" << std::endl;

                    mSaveFile.close();
                    
		    std::string rmcmd = "/usr/bin/rm -f ";
                    rmcmd += mSaveFileName;
                    system(rmcmd.c_str());

                    return;
                } else {
                    mName = tmpname;
                }

                // make sure we have enough buckets allocated for the restore
                // The time doesn't matter because we're just over-writing
                // everything
                while (mQueue.size() < number_saved) 
                    mQueue.push_back(mDummyVal);

                std::copy(std::istream_iterator<T>(mSaveFile),
                          std::istream_iterator<T>(),
                          mQueue.begin());
                
                mSaveFile.close();

                if (mDebug > 0) {
                    std::cout << "TSWindowMod::restore(): completed "
                              << "restore from " << mSaveFileName << std::endl;
         
                    if (mDebug > 2)
                        dump();
                }
            }
        }


    /// Restore from named file
    void restore(const std::string & savefile)
        {
            mSaveFileName = savefile;
            restore();
        }

   
    /// Dump data to stderr
    void dump() const
        {
            if (mDebug > 2) {
                std::cout << "TSWindowMod debug dump:" << std::endl;
                std::cout << "\tmName  = " << name() << std::endl;
                std::cout << "\tmDebug = " << mDebug << std::endl;
                std::cout << "\tmMaxLength = " << maxLength() << std::endl;
                std::cout << "\tmStartTime = " << startTime() << std::endl;
                std::cout << "\tmDt = " << mDt << std::endl;
                std::cout << "\tmSaveFileName = " << saveFileName() << std::endl;
                std::cout << "\tData: " << std::endl;
                std::copy(mQueue.begin(), mQueue.end(),
                          std::ostream_iterator<T>(std::cout, " "));
                std::cout << std::endl;
            }
        }

   
private:
    /// Debug level
    int         mDebug;
   
    /// Name of window
    std::string mName;
   
    /// Max. length of window
    size_t      mMaxLength;
   
    /** The actual deque object. Time ordered, i.e. data gets
     * pushed onto the back
     */
    std::deque<T> mQueue;
   
    /// The start time of a window
    ulong_t     mStartTime;
   
    /// The stride (time step) of window
    ulong_t     mDt;
   
    /// array of data for TSeries
    T          *mData;
   
    /// File for saving data
    std::string  mSaveFileName;
    std::fstream mSaveFile;
   
    /// Value to use as dummy data
    T           mDummyVal;

    /// Smart push_back
    void smart_push_back(T val)
        {
            // if we aren't full
            if (mQueue.size() < mMaxLength) {
                if (mDebug > 1)
                    std::cout << "TSWindowMod::smart_push_back(): pushing "
                              << val << std::endl;
                mQueue.push_back(val);  // later values at back
            } else {
                if (mDebug > 1) {
                    std::cout << "TSWindowMod::smart_push_back(): deque full. "
                              << "Popping front" << std::endl;
                    std::cout << "TSWindowMod::smart_push_back(): pushing "
                              << val << std::endl;
                }
         
                mQueue.pop_front();
                mQueue.push_back(val);
                mStartTime += mDt;
         
                if (mDebug > 1)
                    std::cout << "TSWindowMod::smart_push_back(): mStartTime = "
                              << mStartTime << std::endl;
            }
        }


    /** compute gap between the time of the last data in queue and the
     * time given */
    long long computeGap(const Time & t)
        {
            if (mDebug > 2) {
                std::cout << "TSWindowMod::computeGap(): "
                          << "t.getS() = " << t.getS()
                          << "; mStartTime = " << mStartTime
                          << "; mQueue.size() = " << mQueue.size()
                          << std::endl;
            }

            if (mDt == 0) {
                std::string eObj("TSWindowMod::computeGap(): ERROR: mDt == 0");
                std::cerr << eObj << std::endl;
                throw std::runtime_error(eObj);
            }

            // (given time - (time of last data in deque)) / dt
            return (static_cast<long long>(t.getS()) -
                    (static_cast<long long>(mStartTime) +
                     static_cast<long long>(mQueue.size())*static_cast<long long>(mDt))) /
                static_cast<long long>(mDt);
        }
};

#endif     //  LOCKLOSS_TSWINDOWMOD_HH
