#ifndef SlopeFilter_HH
#define SlopeFilter_HH

#define MAX_CHANNEL_NAME_LENGTH 30

#include <string>
#include "Pipe.hh"
#include "TSeries.hh"

class ostream;
class FSeries;

/** Slope filters are filters (linear or nonlinear) that search
  * for nonzero gradient in data. The order of the filter is the
  * length of time over which the gradient is measured divided by
  * the sampling rate of the data channel. 
  * @memo Ramp Filter Implementation API
  * @author Edward J. Daw
  * @email edaw@lsuligo.phys.lsu.edu
  * @version 1.0; last modified July 9, 2001
  */
class SlopeFilter : public Pipe {

public:

  SlopeFilter* clone(void) const{ return 0; }

/** Default constructor. The filter is declared empty. No 
  * storage for past data is declared
  */
  SlopeFilter(void);

/** Copy constructor.
  */
  SlopeFilter(const SlopeFilter&);

/** Constructor with an order and a channel name declared. The time
  * series input is to set the name of the channel to be filtered and
  * get the sampling rate for that channel.
  * @param order Number of points used to measure data slope/
  * @param tsinit Time series to initialize slope detector filter
  * @memo construct slope filter with a time series and an order.
  */
  SlopeFilter(unsigned order, TSeries tsinit);

/** Constructor with an order and a sampling rate declared. The name
  * of the channel is set to unknown. The sampling rate is set manually
  * @param order Number of points used to measure data slope
  * @param samprate Sampling rate in Hz
  * @memo construct a slope filter with a sampling rate and an order.
  */
  SlopeFilter(unsigned order, double samprate);

/** Default Destructor.
  */
  ~SlopeFilter(void);

/** Set the order and sampling rate for the filter
  * @param order Number of points used to measure data slope
  * @param samprate Data sampling rate
  * @memo set the slope filter order and sampling rate
  */
  void init(unsigned order, double samprate);

  /** Set filter parameters for a predeclared slope filter
   */
  void setFilterParams(unsigned order, TSeries tsinit);

/** Apply the filter to a timeseries.
  * @return Filtered time series.
  * @memo Apply filter to a time series
  */
  TSeries apply(const TSeries& data_in);

/** Check data before filtering. Throws an exception on bad data.
  */
  void dataCheck(const TSeries& data_in) const;

/** Clear history buffers, clear error flag.
  */
  void reset(void);

/** See if history buffers have been created.
  * @return 1 (true) if yes, 0 (false) if no.
  * @memo Check filter for coefficients, history buffers
  */
  bool inUse(void) const;

/** Get start time data being filtered.
  * @return Start time for first data set on which slope detector applied.
  * @memo Get data start time
  */
  Time getStartTime(void) const;

/** Get data current time
  * @return Get time for the last data to which the filter was applied.
  * @memo Get time for the last data to which the filter was applied.
  */
  Time getCurrentTime(void) const;

/** Get filter order.
  * The filter order is the number of data samples over which the
  * gradient of the data is measured. 
  * @return Order of filter. 
  */
  unsigned getOrder(void) const;

private:

  // name of channel to which filter is applied
  std::string _channel_name;
  // sampling rate of channel
  double _sample_rate;
  // true if filter history buffers, channel name and sampling
  // rate have been set
  bool _in_use;
  // order of filter. Defined as number of successive samples 
  // used to estimate gradient.
  unsigned _order;
  // history buffer for slope filter
  double* history_buffer;
  // counter for history buffer
  int _buffer_counter;
  // number of elements in history buffer
  int _buffer_length;
};

#ifndef __CINT__

inline unsigned
SlopeFilter::getOrder(void) const {
  return _order;
}

#endif // define __CINT__
#endif // define SlopeFilter_HH





