
#include <time.h>
#include "events/Events.hh"
#include "Histogram1.hh"
#include "Histogram2.hh"
#include "TSeries.hh"

   using namespace std;
   using namespace events;




   // MyCondition: derive from Condition 
   class MyCondition : public events::Condition { 
   public: 
           // Constructor 
      MyCondition (int n) : mN (n) {
      }; 
           // Override Copy method; define copy constructor if necessary 
      virtual MyCondition* Copy() const { 
         return new MyCondition (*this); } 
           // Override the Evaluate method 
      virtual bool Evaluate (const events::Argument& arg, 
                        bool& val) const; 
   private: 
      int              mN; 
   }; 
   // MyFunction: derive from Function 
   class MyFunction : public events::Function { 
   public: 
           // Default constructor 
      MyFunction() : mC1 ("Amplitude"), mC2 ("Noise") {
      } 
           // Override Copy method; define copy constructor if necessary 
      virtual MyFunction* Copy() const { 
         return new MyFunction (*this); } 
           // Override the Evaluate method 
      virtual bool Evaluate (const events::Argument& arg, 
                        events::Value& val) const; 
   private: 
           // Define the columns we are want to access 
      events::Column   mC1; 
      events::Column   mC2; 
   };

   bool MyCondition::Evaluate (const events::Argument& arg, 
                     bool& val) const 
   { 
      val = (arg.Size() == mN);  // Compute the condition 
      return true;               // successfully evaluated the condition 
   }

   bool MyFunction::Evaluate (const events::Argument& arg, 
                     Value& val) const 
   { 
      // Compute the function 
      Value sum (0.0); 
      for (ConstIterator i = arg.Begin(); i != arg.End(); ++i) { 
         Value amp, noise; 
         if (!mC1.Get(*i, amp) || !mC2.Get(*i, noise)) { 
            return false;  // Getting the column values failed: quit 
         } 
         sum += pow (amp / noise, Value (2)); 
      } 
      val = sqrt (sum); 
      return true;         // successfully evaluated the function 
   }


   int main (int argn, const char* argv[]) 
   {
      // Read some events from file
      Set events ("eventfile.xml");
      // Define an empty event set
      Set selected;
      // Define a second empty event set
      Set selected2;
   
      // Select a all burst events with amplitude > 5
      selected.Select (events, Filter ("burst::*") && Column("Amplitude") > 5);
      // Define a histogram (20 bins from 0 to 20)
      Histogram1 h1 ("H1", 20, 0, 20);
      // Fill it with the amplitude of the selected events
      selected.Histogram (h1, Column("Amplitude"));
   
      // Select another set of events: all Burst which are not vetoed by
      // a PSL glitch
      selected.Clear();
      // Important: use ConditionPtr rather than condition
      ConditionPtr cond_bursts = 
         Filter ("burst::*") && Column("Amplitude") > 5;
      ConditionPtr cond_pslglitch = 
         Filter ("glitch::psl") && (Column("Amplitude") > 10);
      // Set the veto window to -0.5 to +0.5 sec
      selected.Select (events, cond_bursts && !Veto (cond_pslglitch, -0.5, 1.0));
      // Again histogram it
      Histogram1 h2 ("H2", 20, 0, 20);
      selected.Histogram (h2, Column("Amplitude"));
   
      // Or do it all in one
      ConditionPtr cond_vetoedbursts = 
         cond_bursts && !Veto (cond_pslglitch, -0.5, 1.0);
      Histogram1 h3 ("H3", 20, 0, 20);
      events.Histogram (h3, Column("Amplitude"), cond_vetoedbursts);
   
      // Make sure we only have single H1 and L1 events
      events.Select (IfoSet ("H1L1", IfoSet::kAnyOne));
      // select all non-vetoed H1 burst
      selected.Select (events, IfoSet ("H1") && cond_vetoedbursts);
      // add all non-vetoed L1 burst
      selected.SelectAdd (events, IfoSet ("L1") && cond_vetoedbursts);
      // Now look for coincidences; set the coincidence window is 100 ms
      selected.Coincidence (0.1, IfoSet ("H1[0]") && IfoSet ("L1[1]"));
      // Make a 1D histogram of the time difference
      Histogram1 h4 ("H4", 100, -0.1, 0.1);
      selected.Histogram (h4, Column("Time[1]") - Column("Time[0]"));
   
      // Or do it all in one
      ConditionPtr cond_H1 = IfoSet ("H1[0]") && cond_vetoedbursts;
      // Shift the condition to the second event
      ConditionPtr cond_L1 = IfoSet ("L1[1]") && Shift (cond_vetoedbursts);
      events.Histogram (h4, Column("Time[1]") - Column("Time[0]"), 
                       cond_H1 && cond_L1);
   
      // Check up on clusters
      Set clusters;
      clusters.SetWindow (10.0, -5.0);
      clusters.Clusters (2, 10);
      // Plot cluster number against time
      TSeries t1;
      clusters.TimeSeries (t1, Column("N"));
   
      // Algorithm tests
      /////////////////////////////////////////////////////////////////////
   
      // Calculate the significance from the amplitude and sigma
      {
         Set s1 ("input.xml");
         SetColumn (s1.Begin(), s1.End(), 
                   "Significance",
                   Column ("Amplitude") / Column ("Sigma"));
      }
     // Calculate the significance from the amplitude and sigma
     // for all burts events
      {
         Set s1 ("input.xml");
         SetColumn (s1.Begin(), s1.End(), 
                   Column ("Significance"),
                   Column ("Amplitude") / Column ("Sigma"),
                   Filter ("burst::*"));
      }
   
      // Select all events in s1 which have an amplitiude
      // greater than 5 and add them to s2.
      {
         Set s1 ("input.xml");
         Set s2;
         Select (s1.Begin(), s1.End(),
                back_inserter (s2), Column("Amplitude") > 5);
      }
   
      // Remove all non-burst events from the event set.
      {
         Set s1 ("input.xml");
         Iterator rem = Remove (s1.Begin(), s1.End(),
                              !Filter ("burst:*"));
         s1.Erase (rem, s1.End());
      }
   
      // Select all coincidence events in s1 using a coincidence 
      // window of 100ms and add them to s2
      {
         Set s1 ("input.xml");
         Set s2;
         Coincidence (s1.Begin(), s1.End(), 2,
                     back_inserter (s2), 0.1);
      }
   
      // Select all coincidence events in s1 using a coincidence 
      // window of 100ms and add them to s2
      {
         Set s1 ("input1.xml");
         Set s2 ("input2.xml");
         Set s3;
         Coincidence (s1.Begin(), s1.End(),
                     s2.Begin(), s2.End(),
                     back_inserter (s3), 
                     0.1);
      }
      // Select all coincidence events in s1 using a coincidence 
      // window of 100ms and add them to s2
      {
         Set s1 ("input1.xml");
         Set s2 ("input2.xml");
         Set s3;
         Iterator b = s1.Begin();
         Iterator e = s1.End();
         Coincidence (b, e, s2.Begin(), s2.End(),
                     back_inserter (s3), 
                     0.1);
      }
   
      // Select all tripe coincidence events between sets s1, s2
      // and s3 using a coincidence window of 100ms, and add them 
      // to s4. Additionally, we ask that each event has an amplitude
      // greater than 10.
      {
         Set s1 ("input1.xml");
         Set s2 ("input2.xml");
         Set s3 ("input3.xml");
         Set s4;
         Coincidence (s1.Begin(), s1.End(),
                     s2.Begin(), s2.End(),
                     s3.Begin(), s3.End(),
                     back_inserter (s4), 0.1, 
                     Column ("Amplitude[0]") > 10 &&
                     Column ("Amplitude[1]") > 10 &&
                     Column ("Amplitude[2]") > 10);
      }
      // Select all events in s1 which are in coincidence with an event
      // from s2 through s5 using a coincidence window of 100ms,
      // and add them to s6
      {
         Set s1 ("input1.xml");
         Set s2 ("input2.xml");
         Set s3 ("input3.xml");
         Set s4 ("input4.xml");
         Set s5 ("input5.xml");
         Set s6;
         WindowIterator::InputStateList state;
         state.push_back (WindowIterator::InputState (s1.Begin(), s1.End()));
         state.push_back (WindowIterator::InputState (s2.Begin(), s2.End()));
         state.push_back (WindowIterator::InputState (s3.Begin(), s3.End()));
         state.push_back (WindowIterator::InputState (s4.Begin(), s4.End()));
         state.push_back (WindowIterator::InputState (s5.Begin(), s5.End()));
         Coincidence (state, back_inserter (s6), 0.1);
      }
      // Sort all events of s with descending amplitude
   //       {
   //          Set s1 ("input.xml");
   //          Sort (s1.Begin(), s1.End(), Column("Amplitude"), false);
   //       }
   
      // Pick the 20 events with largest amplitude and put
      // the in descending order into s2.
      {
         Set s1 ("input.xml");
         Set s2;
         Sort (s1.Begin(), s1.End(), back_inserter (s2),
              Column("Amplitude"), false, 20);
      }
   
      // Histogram the amplitude of events from set s1 which are
      // within the time interval 677000000 to 678000000.
      {
         Set s1 ("input.xml");
         Histogram1 h1 ("H1", 20, 0, 20);
         MakeHistogram (h1, s1.Begin(), s1.End(), Column("Amplitude"),
                       Time(677000000) <= Column("Time") < Time(678000000));
      }
   
      // Histogram the amplitude of events from set s1 which are
      // within the time interval 677000000 to 678000000 against
      // their time.
      {
         Set s1 ("input.xml");
         Histogram2 h2 ("H2", 100, 0, 100, 100, 0, 1000000);
         MakeHistogram (h2, s1.Begin(), s1.End(), 
                       Column("Time") - Time(677000000),
                       Column("Amplitude"),
                       Time(677000000) <= Column("Time") < Time(678000000));
      }
   
    // Make a time series of the event rate from 677000000 to 677003600.
      {
         Set s1 ("input.xml");
         TSeries ts (Time (677000000, 0), Interval (1.0), 3600);
         MakeTimeSeries (ts, s1.Begin(), s1.End());
      }
      // Make a time series of the event amplitudes within the time 
      // interval 677000000 to 678000000.
      {
         Set s1 ("input.xml");
         TSeries ts (Time (677000000), Interval (100), 10000);
         MakeTimeSeries (ts, s1.Begin(), s1.End(), 
                        Column("Amplitude"),
                        Time(677000000) <= Column("Time") < Time(678000000));
      }
   
      // Look for a coincidence event
      Set set ("input.xml");
      ConditionPtr cond = IfoSet ("H1[0]") && IfoSet ("L1[1]");
      WindowIterator i (set.Begin(), set.End(), 2);
      WindowIterator end (set.End(), set.End(), 2);
      for (; i != end; ++i) {
         bool ret;
         if (cond->Evaluate (*i, ret) && ret) {
          // found a coincidence event between H1 and L1
         }
      }
   
      // Test home grown function and condition
      {
         Set s3 ("test.xml"); 
         Histogram1 h1 ("Test", -1, +1, 200); 
         s3.Histogram (h1, MyFunction(), MyCondition(3));
      }
   
      return 0;
   }

