/* -*- mode: c++; c-basic-offset: 4; -*- */
#ifndef TABLE_TABLE_HH
#define TABLE_TABLE_HH

#include "column.hh"

//======================================  Table class definition
namespace table {

  /** Table manipulation class.
   */
  class Table {
  public:
    typedef unsigned int col_index;
    enum sort_enum {kAuto, kMerge, kBubble};
  public:
    Table(void) : mRow(0), mDebug(0) {}
    Table(const std::string& title, const std::string& in);

    /**  Check if table is defined.
      *  @memo Is table defined
      *  @return true if table structure is defined.
      */
    bool defined(void) const;

    /**  Check if table has any rows
      *  @memo Does table have rows?
      *  @return true if columns are empty.
      */
    bool empty(void) const;

    /**  Check if column is defined.
      *  @memo Does column exist
      *  @param name Column name.
      *  @return true if column is defined.
      */
    bool exists(const std::string& name) const;

    /**  Get the index of the row with the specified table. If the 
      *  specified column name is not found, the value returned is 
      *  equal to the table size (i.e. the number of columns).
      *  @memo Get column index
      *  @param name name of column to be found
      *  @return Column index
      */
    col_index find(const std::string& name) const;

    Column& findColumn(const std::string& name);
    const Column& findColumn(const std::string& name) const;
    const char* getTitle(void) const;
    Column::row_index getNRows(void) const;
    void insertColumn(col_index b4, const std::string& title);
    void insertRows(Column::row_index b4, const Table& tab);
    void merge(col_index iCol, col_index jCol, const std::string& join);
    void move_row(Column::row_index i, Column::row_index j);
    void read(const std::string& in);
    void readtxt(const std::string& in);
    void readxml(const std::string& in);
    void removeColumn(col_index col);
    void selectRows(const std::vector<bool>& mask);
    void setDebug(int i);
    col_index size(void) const;
    void sort(col_index col, bool decrease=false, sort_enum t=kAuto);
    void swap_rows(Column::row_index i, Column::row_index j);
      /**  Write out the table to the specified file name. The output format
        *  is detemined by the file name extension. Text files contain 
	*  information specified by the flag word: a header (1), the table 
	*  body (2) and a summary line (4).
	*  @memo write the table to a specified file.
	*  @param f The output file name.
	*  @param flg Word containing Ored flags.
	*  @param t title?
	*/
    void write(const std::string& f, int flg=7, const std::string& t=0) const;
    std::ostream& writehtml(std::ostream& out) const;
    std::ostream& writetxt(std::ostream& out, int flags=7) const;
    std::ostream& writexml(std::ostream& out) const;
    Column& operator[](col_index col);
    const Column& operator[](col_index col) const;
  private:
    void bsort(col_index col, bool decrease);
    void bsort(std::vector<double> index, bool decrease);
    void msort(col_index col, bool decrease);
    void msort(std::vector<double> index, bool decrease);
  private:
    Column::row_index mRow;
    std::string mTitle;
    std::vector<Column> mTable;
    int mDebug;
  };

  inline bool
  Table::defined(void) const {
      return !mTable.empty();
  }

  inline bool
  Table::empty(void) const {
      return !mRow;
  }

  inline Table::col_index
  Table::size(void) const {
      return mTable.size();
  }

  inline bool
  Table::exists(const std::string& name) const {
      return find(name) != mTable.size();
  }

  inline Column::row_index 
  Table::getNRows(void) const {
      return mRow;
  }

  inline const char* 
  Table::getTitle(void) const {
      return mTitle.c_str();
  }

  inline Column& 
  Table::operator[](col_index col) {
      return mTable[col];
  }

  inline const Column& 
  Table::operator[](col_index col) const {
      return mTable[col];
  }

}
#endif // !defined(TABLE_TABLE_HH)
